"""
IMX Flash Control Block module

This module parses and allow manipulation of the Flash Control Block.

@author Damien Cauquil <damien.cauquil@digital.security>
"""

from math import ceil
from struct import unpack
from termcolor import colored

def formatval(v):
    """
    Format value string.
    """
    return colored(v, 'cyan', attrs=['bold'])

def u32le(b, offset):
    """
    Reads a LE unsigned integer from `b` at the given `offset`.
    """
    return unpack('<I', b[offset:offset+4])[0]


class FCBError(Exception):
    """
    FCB Error descriptor
    """

    def __init__(self, message):
        """
        @param  message string  Message to associate with this error.
        """
        Exception.__init__(self)
        self.message = message

    def __repr__(self):
        """
        Class representation
        """
        return '<FCBError msg="%s">' % self.message


class FCB(object):
    """
    Flash Control Block class
    """

    def __init__(self, content):
        """
        Constructor, parses the FCB content.
        """
        if len(content) >= 132:
            try:
                # Unpack all the fields
                self.magic = content[4:8]
                if self.magic != b'FCB ':
                    raise FCBError('Wrong FCB magic value (%08x instead of %08x)' % (
                        unpack('<I', self.magic),
                        unpack('<I', b'FCB ')
                    ))
                self.version = unpack('>I', content[8:12])[0]
                self.page_data_size = u32le(content, 20)
                self.total_page_size = u32le(content, 24)
                self.sectors_per_block = u32le(content, 28)
                self.nb_nands = u32le(content, 32)
                self.ecc_block_type = u32le(content, 44)
                self.ecc_block0size = u32le(content, 48)
                self.ecc_blockNsize = u32le(content, 52)
                self.ecc_block0type= u32le(content, 56)
                self.metadata_bytes = u32le(content, 60)
                self.nb_ecc_blocks_per_page = u32le(content, 64)
                self.fw1_start = u32le(content, 104)
                self.fw2_start = u32le(content, 108)
                self.pages_fw1 = u32le(content, 112)
                self.pages_fw2 = u32le(content, 116)
                self.bch_type = u32le(content, 136)
                self.bb_marker = u32le(content, 124)
                self.bb_marker_bits = u32le(content, 128)

                # We compute the raw marker offset
                self.marker_page = int(self.bb_marker/self.ecc_blockNsize)
                self.marker_offset = self.bb_marker % self.ecc_blockNsize
                self.marker_raw_offset = int(self.metadata_bytes + ceil((self.ecc_block0type*26)/8) + ceil((self.marker_page-1)*(self.ecc_block_type*26)/8) + self.bb_marker)
            except TypeError as type_exc:
                raise FCBError("Content MUST be bye of type `bytes`")
        else:
            raise FCBError("FCB content must contain at least 132 bytes")

    def get_ecc_block0_size(self):
        """
        Get the first block ECC size in bits.

        @return int ECC size in bits
        """
        return (self.ecc_block0type * 26)

    def get_ecc_block0_strength(self):
        """
        Get the first block ECC strength (BCH).

        @return int ECC strength (BCH)
        """
        return (self.ecc_block0type)

    def get_ecc_blockN_size(self):
        """
        Get the next blocks ECC size in bits.

        @return int ECC size in bits
        """
        return (self.ecc_block_type * 26)

    def get_ecc_blockN_strength(self):
        """
        Get the next blocks ECC strength (BCH).

        @return int ECC strength (BCH)
        """
        return (self.ecc_block_type)

    def get_data_block0_size(self):
        """
        Get first block data size on which the ECC is computed.

        @return int data block size
        """
        return self.ecc_block0size

    def get_data_blockN_size(self):
        """
        Get next blocks data size on which the ECC is computed.

        @return int data block size
        """
        return self.ecc_blockNsize

    def set_metadata_bytes(self, metadata_bytes):
        """
        Force metadata bytes size.

        @param metadata_bytes   int Metadata bytes size
        """
        self.metadata_bytes = metadata_bytes

    def set_page_data_size(self, page_data_size):
        """
        Set page data size.

        @param page_data_size   int Page size in bytes
        """
        self.page_data_size = page_data_size

    def set_ecc_size(self, ecc_size):
        """
        Set ECC size in bits.

        @param ecc_size int ECC size in bits.
        """
        self.ecc_block0type = int(ecc_size/26)
        self.ecc_block_type = int(ecc_size/26)

    def set_bb_marker(self, bb_marker):
        """
        Force bad block marker offset.

        @param bb_marker chr Bad block marker offset to use.
        """
        self.bb_marker = bb_marker
        # We compute the raw marker offset
        self.marker_page = int(self.bb_marker/self.ecc_blockNsize)
        self.marker_offset = self.bb_marker % self.ecc_blockNsize
        self.marker_raw_offset = int(self.metadata_bytes + ceil((self.ecc_block0type*26)/8) + ceil((self.marker_page-1)*(self.ecc_block_type*26)/8) + self.bb_marker)

    def display(self):
        """
        Display FCB to standard output.
        """
        # Display information
        print('FCB version: %d' % self.version)
        print('')
        print('---[ NAND structure ]---------')
        print(' > Page data size:\t'+ formatval('%d bytes'% self.page_data_size))
        print(' > Total page size:\t' + formatval('%d bytes (OOB: %d bytes)' % (self.total_page_size, self.total_page_size - self.page_data_size)))
        print(' > Sectors/block:\t' + formatval('%d' % self.sectors_per_block))
        print(' > Number of Nands:\t' + formatval('%d' % self.nb_nands))
        print('')
        print('---[ ECC ]--------------------')
        print(' > ECC block 0 type:\t' + formatval('%d (%d bits)' % (self.ecc_block0type, self.ecc_block0type*26)))
        print(' > ECC block 0 size:\t' + formatval('%d bytes' % self.ecc_block0size))
        print(' > ECC block N type:\t' + formatval('%d (%d bits)' % (self.ecc_block_type, self.ecc_block_type*26)))
        print(' > ECC block N size:\t' + formatval('%d bytes' % self.ecc_blockNsize))
        print(' > Metadata bytes:\t' + formatval('%d' % self.metadata_bytes))
        # Original NumEccBlocksPerPage does not include Block0
        print(' > ECC blocks/page:\t' + formatval('%d' % (self.nb_ecc_blocks_per_page +1)))
        print(' > ECC BCH Type:\t%d' % self.bch_type)
        print('')
        print('---[ BadBlocks ]--------')
        print(' > Bad block marker byte:\t' + formatval('0x%x' % self.bb_marker))
        print(' > Bad block start bit:\t\t' + formatval('0x%x' % self.bb_marker_bits))
        print(' > Bad block Marker raw offset:\t' + formatval('0x%x' % self.marker_raw_offset))
        print('')
        print('---[ Firmware Info]-----')
        print(' > Firmware #1:\t' + formatval('start @%08x (%d pages, %d bytes)' % (self.fw1_start, self.pages_fw1, self.pages_fw1*self.page_data_size)))
        print(' > Firmware #2:\t' + formatval('start @%08x (%d pages, %d bytes)' % (self.fw2_start, self.pages_fw2, self.pages_fw1*self.page_data_size)))



"""
IMX Tools main module

@author Damien "virtualabs" Cauquil <damien.cauquil@digital.security>
"""
import sys
import os.path
import bchlib
import progressbar

from math import ceil
from struct import unpack
from argparse import ArgumentParser

from termcolor import colored
from imxtools.fcb import FCB, FCBError

def skip_bits(page, nbits, nbytes):
    """
    Skip bits for a given page
    """
    # compute relative shifting and complementary shifting values
    rel_shift = int(nbits%8)
    comp_shift= int(8 - rel_shift)

    # skip nbits/8 bytes first
    page = page[int(nbits/8):]
    page += bytes([0])

    if rel_shift > 0:
        # Loop over bytes and shift rel_shift bits to the left
        output = []
        for i in range(nbytes):
            output.append(
                (page[i]>>rel_shift) | ((page[i+1] << comp_shift)&0xff)
            )
    else:
        output = page
    return output

def parse_fcb(content, verbosity=None, display=False):
    """
    Parse and display the FCB structure.
    """
    fcb = FCB(content)

    # Display FCB if required
    if display:
        fcb.display()

    # return FCB info
    return fcb


def ecc_correct(block, code, ecc_strength):
    """
    Try to fix `block` with `code`, for an ECC size of `bitsize`.
    """
    ecc = bchlib.BCH(8219, ecc_strength, reverse=True)
    try:
        block_ = ecc.correct(bytes(block), bytes(code))
        return block_
    except Exception as e:
        return block


def process_page(page, fcb, ecc=False):
    """
    Split page in blocks and ecc codes.
    """
    # First, remove metadata bytes
    page_size = len(page)
    marker = page[0]
    assert(page[fcb.marker_raw_offset]==0xff)
    page = page[:fcb.marker_raw_offset] + bytes([marker]) + page[fcb.marker_raw_offset+1:]
    page = page[fcb.metadata_bytes:]

    blocks = []
    nb_blocks = fcb.nb_ecc_blocks_per_page+1

    # Iterate over each block
    n = 8
    for i in range(nb_blocks):
        # First block is processed separately
        if i==0:
            ecc_size = fcb.get_ecc_block0_size()
            ecc_strength = fcb.get_ecc_block0_strength()
            block_size = fcb.get_data_block0_size()
        else:
            ecc_size = fcb.get_ecc_blockN_size()
            ecc_strength = fcb.get_ecc_blockN_strength()
            block_size = fcb.get_data_blockN_size()

        ecc_nb_bytes = int(ecc_size/8)
        if (fcb.get_ecc_block0_size()%8 > 0):
            ecc_nb_bytes += 1
        ecc = page[block_size:block_size+ecc_nb_bytes]


        # copy block_size bytes
        block = page[:block_size]

        # try to correct block if required
        if ecc:
            block = ecc_correct(block, ecc, ecc_strength*2)

        # save block
        blocks.append(block)

        # skip ecc_size (in bits) bits
        page = skip_bits(page, block_size*8 + ecc_size, int(((block_size*8 + ecc_size)*(n-1))/8))
        n -= 1

    # Align to original page size
    output = []
    for block in blocks:
        output.extend(block)
    return bytes(output)

def extract_firmware(content, fcb, output, firmware_id, bb_marker_override=None, metadata_override=None, page_size_override=None, correct_ecc=False, ecc_size_override=None):
    """
    Convert an IMX NAND dump to memory dump.
    """
    # Apply overrides if defined
    if metadata_override is not None:
        fcb.set_metadata_bytes(metadata_override)
    if page_size_override is not None:
        fcb.set_page_data_size(page_size_override)
    if ecc_size_override is not None:
        fcb.set_ecc_size(ecc_size_override)
    if bb_marker_override is not None:
        fcb.set_bb_marker(bb_marker_override)

    # We keep only a fraction of the content, as described in the FCB
    if firmware_id == 1:
        content = content[fcb.fw1_start * fcb.total_page_size:(fcb.fw1_start + fcb.pages_fw1) * fcb.total_page_size]
    elif firmware_id == 2:
        content = content[fcb.fw2_start * fcb.total_page_size:(fcb.fw2_start + fcb.pages_fw2) * fcb.total_page_size]

    blocksize = fcb.total_page_size
    globsize = len(content)
    nbblocks = int(globsize/blocksize)

    valid_blocks = 0
    bad_blocks = 0
    output = open(output,'wb')
    bar = progressbar.ProgressBar(max_value=nbblocks)
    for i in range(nbblocks):
        bar.update(i)
        page = content[i*blocksize:(i+1)*blocksize]
        c_block = process_page(page, fcb, correct_ecc)
        output.write(c_block)
    output.close()

def convert_nand_dump(content, fcb, output, bb_marker_override=None, metadata_override=None, page_size_override=None, correct_ecc=False, ecc_size_override=None):
    """
    Convert an IMX NAND dump to memory dump.
    """
    # Apply overrides if defined
    if metadata_override is not None:
        fcb.set_metadata_bytes(metadata_override)
    if page_size_override is not None:
        fcb.set_page_data_size(page_size_override)
    if ecc_size_override is not None:
        fcb.set_ecc_size(ecc_size_override)
    if bb_marker_override is not None:
        fcb.set_bb_marker(bb_marker_override)

    blocksize = fcb.total_page_size
    globsize = len(content)
    nbblocks = int(globsize/blocksize)

    valid_blocks = 0
    bad_blocks = 0
    output = open(output,'wb')
    bar = progressbar.ProgressBar(max_value=nbblocks)
    for i in range(nbblocks):
        bar.update(i)
        page = content[i*blocksize:(i+1)*blocksize]
        c_block = process_page(page, fcb, correct_ecc)
        output.write(c_block)
    output.close()


def find_fcb_offset(content):
    """
    Find Flash Control Block offset

    @return int offset >=0 on success, -1 if an error occured.
    """
    try:
        index = content.index(b'FCB ')
        if (index > 4):
            return (index-4)
        return -1
    except ValueError as exc:
        return -1



"""
         ___ __  ____  __  _____         _
        |_ _|  \/  \ \/ /_|_   _|__  ___| |___
         | || |\/| |>  <___|| |/ _ \/ _ \ (_-<
        |___|_|  |_/_/\_\   |_|\___/\___/_/__/

IMX Nand Information tool
=========================

This tool


"""

from argparse import ArgumentParser
from imxtools import parse_fcb, find_fcb_offset
from termcolor import colored

def main():
    parser = ArgumentParser()
    parser.add_argument('-v', '--verbose', action='store_true', default=False, dest='verbose', help='Be more verbose')
    parser.add_argument('-o', '--offset', dest='offset', help='Force FCB offset value')
    parser.add_argument('nand_dump', type=str, help='Raw NAND dump to process')
    args = parser.parse_args()

    print(colored('''
     ___ __  ____  __  _____         _
    |_ _|  \/  \ \/ /_|_   _|__  ___| |___
     | || |\/| |>  <___|| |/ _ \/ _ \ (_-<
    |___|_|  |_/_/\_\   |_|\___/\___/_/__/

      ---< IMX Nand Info >---
    ''', 'cyan', attrs=['bold']))

    # Load file
    if args.verbose:
        print('>> Loading memory dump ...')
    dump = open(args.nand_dump, 'rb').read(4096)

    # Override offset if provided
    if args.offset is not None:
        # Accept decimal and hexadecimal offset values
        if args.offset.lower().startswidth('0x'):
            offset = int(args.offset, 16)
        else:
            offset = int(args.offset)
        if args.verbose:
            print('>> Forcing FCB offset to %x' % offset)
    else:
        offset = find_fcb_offset(dump)

    if offset < 0:
        print('!!'+colored('FCB not found, check your dump.', 'red', attrs=['bold']))
    else:
        if args.verbose:
            print('>> Dumping FCB details ...')
        # Display NAND info
        fcb = dump[offset:offset+140]
        parse_fcb(fcb, verbosity=args.verbose, display=True)



import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "imx-nand-tools",
    version = "1.0.0",
    url='https://github.com/DigitalSecurity/imx-nand-tools',
    author = "Damien Cauquil",
    author_email = "damien.cauquil@digital.security",
    description = ("Freescale i.MX NAND reverse tools"),
    license = "MIT",
    keywords = "imx freescale tool",
    packages=['imxtools'],
    install_requires=[
            'progressbar2',
            'termcolor',
            'bchlib'
    ],
    entry_points= {
        'console_scripts': [
            'imx-nand-info=imxtools.imx_nand_info:main',
            'imx-nand-convert=imxtools.imx_nand_convert:main'
        ]
    },
)



"""
         ___ __  ____  __  _____         _
        |_ _|  \/  \ \/ /_|_   _|__  ___| |___
         | || |\/| |>  <___|| |/ _ \/ _ \ (_-<
        |___|_|  |_/_/\_\   |_|\___/\___/_/__/

IMX Nand Conversion tool
=========================

"""

from argparse import ArgumentParser
from imxtools import convert_nand_dump, find_fcb_offset, extract_firmware
from imxtools.fcb import FCB
from termcolor import colored


def main():
    parser = ArgumentParser()
    parser.add_argument('-o', '--offset', dest='offset', help='Force FCB offset value')
    parser.add_argument('-b', '--bad-block-offset', dest='bb_offset', help='Force bad block marker offset')
    parser.add_argument('-p', '--page-size', dest='page_size', help='Force page size (in bytes)')
    parser.add_argument('-m', '--metadata-size', dest='metadata_size', help='Force metadata size (in bytes)')
    parser.add_argument('-e', '--ecc-size', dest='ecc_size', help='Force ECC size (in bits)')
    parser.add_argument('-v', '--verbose', action='store_true', default=False, dest='verbose', help='Be more verbose')
    parser.add_argument('-f', '--firmware', dest='firmware', type=int, help='Firmware number to extract (default: 1)')
    parser.add_argument('-c', '--correct', dest='ecc', action='store_true', default=False, help='Correct errors with ECC')
    parser.add_argument('nand_dump', type=str, help='Raw NAND dump to process')
    parser.add_argument('output_nand_dump', type=str, help='Output NAND dump')
    args = parser.parse_args()

    print(colored('''
     ___ __  ____  __  _____         _
    |_ _|  \/  \ \/ /_|_   _|__  ___| |___
     | || |\/| |>  <___|| |/ _ \/ _ \ (_-<
    |___|_|  |_/_/\_\   |_|\___/\___/_/__/

      ---< IMX Nand Convert >---
    ''', 'cyan', attrs=['bold']))

    # Load file
    if args.verbose:
        print('>> Loading memory dump ...')
    dump = open(args.nand_dump, 'rb').read()

    # Override offset if provided
    if args.offset is not None:
        # Accept decimal and hexadecimal offset values
        if args.offset.lower().startswidth('0x'):
            offset = int(args.offset, 16)
        else:
            offset = int(args.offset)
    else:
        offset = find_fcb_offset(dump)

    if offset < 0:
        print('!!'+colored('FCB not found, check your dump.', 'red', attrs=['bold']))
    else:
        if args.firmware is not None:
            if args.firmware in [1,2]:
                print('>> FCB found at offset 0x%08x' % offset)
                print('>> Extracting firmware #%d ...' % args.firmware)

                # Parse FCB
                fcb = FCB(dump[offset:offset+140])

                extract_firmware(dump, fcb, args.output_nand_dump, args.firmware, args.bb_offset, args.metadata_size, args.page_size, args.ecc, args.ecc_size)
            else:
                print(colored('>> Firmware index MUST be 1 OR 2.', 'red', attrs=['bold']))
        else:
            print('>> FCB found at offset 0x%08x' % offset)
            print('>> Converting image ...')
            # Display NAND info
            fcb = FCB(dump[offset:offset+140])
            convert_nand_dump(dump, fcb, args.output_nand_dump, args.bb_offset, args.metadata_size, args.page_size, args.ecc, args.ecc_size)
